try:
    from fmu_manipulation_toolbox.__version__ import __doc__ as __version__
except ModuleNotFoundError:
    __version__ = "0.0.dev0"

__author__ = "Nicolas.LAURENT@Renault.com"
__copyright__ = "Copyright 2023-2024, Renault SAS"
__license__ = """This code is released under the 2-Clause BSD license. 
See https://github.com/grouperenault/fmu_manipulation_toolbox/blob/main/LICENSE.txt"""
